/***************************************************************************
                          main.c  -  vm86d main file
                             -------------------
    begin                : Mon Mar 31 00:35:52 PHT 2003
    copyright            : (C) 2003 by Antonino Daplas
    email                : adaplas@pol.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/io.h>
#include <linux/vm86.h>
#include <linux/ioctl.h>
#include "lrmi.h"
#include "vbe.h"

static int fd = -1;

static void *LRMI_buffer = NULL;
static struct real_buffer {
	struct vm86_request *request;
	void *data;
} vm86_buffer;


static void vm86_exit(int sig)
{
	int enable = 0;

	if (LRMI_buffer)
		LRMI_free_real(LRMI_buffer);
	if (fd != -1) 
		close(fd);
	exit(sig);
}

static int daemonize(void) {
	switch(fork()) {
	case -1:
		return -1;
	case 0:
		/* child */
		break;
	default:
		/* parent */
		exit(EXIT_SUCCESS);
	}

	setsid();
	umask(0);
	chdir("/");

	return 0;
}

static int process_generic(struct real_buffer *buffer)
{
	LRMI_regs r = *((LRMI_regs *) &buffer->request->regs);

	if (!LRMI_int(buffer->request->interrupt, &r))
		return -1;
	buffer->request->regs = *((struct user_vm86_regs *) &r);
	return 0;
}

static void pre_process_vbe_buffers(struct real_buffer *buffer)
{
	struct vm86_request *request = buffer->request;

	switch (request->regs.eax & 0xff) {
	case 0x00:
	case 0x01:
	case 0x02:
	case 0x09:
	case 0x15:
		if ((request->regs.eax & 0xff) == 0x15 &&
		    !request->regs.ebx)
			break;
		request->regs.es = (unsigned) buffer->data >> 4;
		request->regs.edi = (unsigned)buffer->data & 0xf;
		if (!request->buf_len)
			request->buf_len = 512;
		break;
	case 0x04:
		request->regs.es = (unsigned) buffer->data >> 4;
		request->regs.ebx = (unsigned) buffer->data & 0xf;
		if (!request->buf_len)
			request->buf_len = 512;
		break;
	default:
		break;
	}
}

static void post_process_vbe_buffers(struct real_buffer *buffer,
				     LRMI_regs *r)
{
	struct vm86_request *request = buffer->request;

	switch (r->eax & 0xff) {
	case 0x00:
	{
		struct VBEInfoBlock *vbe = (struct VBEInfoBlock *) buffer->data;
		short *mode_list;
		short *mode_dst = buffer->data + request->buf_len;
		int i = 0;
		if ((request->regs.eax & 0xffff) != 0x4f ||
		    strncmp(vbe->VBESig, "VESA", 4))
			return;
		mode_list = (short *) (vbe->VBEModePtrSeg * 16 +
			vbe->VBEModePtrOff);

		vbe->VBEOEMStringOff = 0; /* don't fill */
		vbe->VBEOEMStringSeg = 0; /* don't fill */
		vbe->VBEOEMVendorOff = 0;
		vbe->VBEOEMVendorSeg = 0;
		vbe->VBEOEMProductOff = 0;
		vbe->VBEOEMProductSeg = 0;
		vbe->VBEOEMProductRevisionOff = 0;
		vbe->VBEOEMProductRevisionSeg = 0;
		vbe->VBEModePtrSeg = 0;
		vbe->VBEModePtrOff = request->buf_len;
		while (*mode_list != -1 && i < (1024 - request->buf_len/2)) {
			*mode_dst++ = *mode_list++;
			i++;
		}
		*mode_dst = -1;
		request->buf_len += (i+1)*2;
		break;
	}
	default:
		break;
	}
}

static int process_vbe(struct real_buffer *buffer)
{
	LRMI_regs r = *((LRMI_regs *) &buffer->request->regs);

	pre_process_vbe_buffers(buffer);
	if (process_generic(buffer))
		return -1;
	post_process_vbe_buffers(buffer, &r);
	return 0;
}

static int process_video(struct real_buffer *buffer)
{
	int retval = -1;
	switch (buffer->request->regs.eax & 0xff00) {
	case 0x4f00:
		retval = process_vbe(buffer);
		break;
	default:
		retval = process_generic(buffer);
		break;
	}
	return retval;
}

static int process_request(struct real_buffer *buffer)
{
	int retval = -1;

	switch (buffer->request->interrupt) {
	case 0x10:
		retval = process_video(buffer);
		break;
	default:
		break;
	}
	return retval;
}

int main()
{
	fd_set set;
	struct timeval timeout;
	int count, enable = 1;

	fd = open("/dev/vm86", O_RDWR | O_ASYNC);
	if (fd == -1) {
		printf("cannot open /dev/vm86\n");
		vm86_exit(1);
	}

	if (!LRMI_init())
		vm86_exit(1);

	if (daemonize())
		vm86_exit(1);

	ioperm(0, 1024, 1);
	iopl(3);

	LRMI_buffer = LRMI_alloc_real(1024 + sizeof(struct vm86_request));
	if (LRMI_buffer == NULL)
		vm86_exit(1);
	vm86_buffer.request = (struct vm86_request *) LRMI_buffer;
	vm86_buffer.data = LRMI_buffer + sizeof(struct vm86_request);
	signal(SIGINT, vm86_exit);
	signal(SIGQUIT, vm86_exit);
	signal(SIGTERM, vm86_exit);
	signal(SIGKILL, vm86_exit);
	signal(SIGPIPE, SIG_IGN);

	/* Initialize the file descriptor set. */
	FD_ZERO (&set);
	FD_SET (fd, &set);

	/* Initialize the timeout data structure. */
	timeout.tv_sec = 5;
	timeout.tv_usec = 0;

	while (1) {
		if (select (FD_SETSIZE, &set, NULL, NULL, NULL)) {
			count = read(fd, LRMI_buffer, 1024);
			if (!count)
				continue;

			count = sizeof(struct vm86_request);
			if (process_request(&vm86_buffer)) {
				memset(vm86_buffer.request, 0, count);
			} else {
				count += vm86_buffer.request->buf_len;
			}
			if (select (FD_SETSIZE, NULL, &set, NULL, NULL)) {
				count = write(fd, LRMI_buffer, count);
			}
		}
	}
	vm86_exit(0);
}
